##
##  src2make.pl -- Generate OpenPKG Makefiles for building/installing packages
##  Copyright (c) 2000-2006 OpenPKG Foundation e.V. <http://openpkg.net/>
##  Copyright (c) 2000-2006 Ralf S. Engelschall <http://engelschall.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;
use OpenPKG::Ctx;
use Data::Dumper;
use IO;
use IO::File;
use Getopt::Long;
use OSSP::cfg;
use DBI;
use DBD::SQLite;

#   configure optional debugging
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   determine OpenPKG information
my $ctx = new OpenPKG::Ctx;
my $progname = $ctx->cmdprog() . " " . $ctx->cmdname();
my $prefix = $ctx->prefix();

#
#   parameters (defaults)
#

my $verbose = 0;
my $help    = 0;
my $debug   = 0;
my $config      = "$prefix/etc/openpkg-tools/bf.conf";
my $prefix      = "";
my $tmpdir      = ($ENV{TMPDIR} || $ENV{TEMPDIR} || "/tmp") . "/openpkg";
my $srcdir      = "";
my $outdir      = "";
my $nouninstall = 0;
my @classes     = "+CORE+BASE+PLUS-EVAL-JUNK";

#
#   utility functions
#

my @cleanup = ();

sub cleanup_remember {
    my ($cmd) = @_;
    push(@cleanup, $cmd);
}

sub cleanup_perform {
    foreach my $cmd (@cleanup) {
        &runcmd($cmd);
    }
}

$SIG{__DIE__} = sub {
    my ($err) = @_;
    $err =~ s|\s+at\s+.*||s if (not $verbose);
    print STDERR "$progname:ERROR: $err ". ($! ? "($!)" : "") . "\n";
    &cleanup_perform() if (not $verbose);
    exit(1);
};

sub verbose {
    my ($msg) = @_;
    print "$msg\n" if ($verbose);
}

sub runcmd {
    my ($cmd) = @_;
    #&verbose("\$ $cmd");
    return system($cmd);
}

sub fullpath {
    my ($prog) = @_;
    my $fullprog = '';
    foreach my $path (split(/:/, $ENV{PATH})) {
        if (-x "$path/$prog") {
            $fullprog = "$path/$prog";
            last;
        }
    }
    return $fullprog;
}

#
#   command line parsing
#

Getopt::Long::Configure("bundling");
my $result = GetOptions(
    'h|help'        => \$help,
    'v|verbose'     => \$verbose,
    'd|debug'       => \$debug,
    'p|prefix=s'    => \$prefix,
    't|tmpdir=s'    => \$tmpdir,
    's|srcdir=s'    => \$srcdir,
    'o|outdir=s'    => \$outdir,
    'n|nouninstall' => \$nouninstall,
    'c|classes=s'   => \@classes,
    'C|config=s'    => \$config
);
if ($help) {
    print "Usage: $progname [options]\n" .
          "Available options:\n" .
          " -h,--help          print out this usage page\n" .
          " -v,--verbose       enable verbose run-time mode\n" .
          " -d,--debug         enable debugging mode\n" .
          " -p,--prefix=PATH   filesystem path to OpenPKG instance\n" .
          " -t,--tmpdir=PATH   filesystem path to temporary directory\n" .
          " -s,--srcdir=PATH   filesystem path to source directory (for source RPMs)\n" .
          " -o,--outdir=PATH   filesystem path to output directory (for Makefiles)\n" .
          " -c,--classes=+A-B  include/exclude class (CORE, BASE, PLUS, PRIV, EVAL, JUNK)" .
          " -C|--config=FILE   specifiy config file ($config)\n";
    exit(0);
}

#   handle config
my $cfg = &readconfig($config);
my $dbfile      = $cfg->{"master"}->{"dbfile"}      || die "mandatory setting master.dbfile undefined";
my $lockretries = $cfg->{"master"}->{"lockretries"} || 250;
my $rpm         = $cfg->{"master"}->{"rpm"}         || "$prefix/bin/openpkg rpm";
my $rpm2cpio    = $cfg->{"master"}->{"rpm2cpio"}    || "$prefix/bin/openpkg rpm2cpio";
my $cpio        = $cfg->{"master"}->{"cpio"}        || "cpio";

$rpm =      &fullpath($rpm)      unless ($rpm      =~ m|/|);
$rpm2cpio = &fullpath($rpm2cpio) unless ($rpm2cpio =~ m|/|);
$cpio =     &fullpath($cpio)     unless ($cpio     =~ m|/|);

$outdir = $srcdir if ($outdir eq '');
if ($prefix eq '' or $srcdir eq '') {
    die "at least --prefix and --srcdir have to be specified";
}
if (not -d $tmpdir) {
    mkdir($tmpdir, 0700) || die "cannot create temporary directory '$tmpdir'";
    &cleanup_remember("rm -rf $tmpdir");
}

#   open and initialize database
my $dbh = &dbinit($dbfile);

my @classcheck;
foreach my $item (@classes) {
    my $remainder = $item;
    my ($classinit, $classhave);
    while ($remainder) {
        $remainder =~ m|^(.)(\w+)(.*)$|;
        ($classinit, $classhave, $remainder) = ($1,$2,$3);
        if ($classinit eq "+") {
            $classinit = 1;
        }
        elsif ($classinit eq "-") {
            $classinit = 0;
        }
        else {
            die "class \"$classhave\" in \"$item\" must be prefixed with + or - but found $classinit";
        }
        push @classcheck, [ $classhave, $classinit ];
    }
}

#
#   scan source directory for Source RPMs and build dependency graphs
#   Examples: foo-1.2-1.0.4.src.rpm
#             foo-1.2-20011121.3.src.rpm
#

my $PGV = {}; # provides-graph for virtual packages
my $DGB = {}; # dependency-graph for building
my $DGI = {}; # dependency-graph for installing
my $PMV = {}; # package maximum/latest version

my @SRC = sort(glob("$srcdir/*.src.rpm"));
&verbose(sprintf("found %d source RPMs under %s", $#SRC+1, $srcdir));
@SRC = () if ($debug and -f "$progname.db");
SRCRPM: foreach my $srcrpm (@SRC) {
    $srcrpm =~ s|^$srcdir/||s;

    &verbose("scanning $srcdir/$srcrpm");

    #   parse source RPM filename into components
    my ($srcrpm_name, $srcrpm_version, $srcrpm_release);
    if (not (($srcrpm_name, $srcrpm_version, $srcrpm_release) =
             ($srcrpm =~ m/^(.+?)-([^-]+)-(\d+\.\d+\.\d+|\d+\.\d{8}|\d{8})\.src\.rpm$/))) {
        die "malformed source RPM filename '$srcrpm'";
    }

    #   query source RPM package
    my $q = '';
    foreach my $t (qw(Name Version Release Distribution Class SigMD5)) {
        $q .= "S-$t:<%{$t}>\n";
    }
    $q .= "[M-BuildPreReq:<%{REQUIRENAME} %|REQUIREFLAGS?{%{REQUIREFLAGS:depflags} %{REQUIREVERSION}}:{}|>\n]";
    my $o = `$rpm -qp --qf "$q" $srcdir/$srcrpm 2>/dev/null`;
    $o =~ s|M-BuildPreReq:<rpmlib\(.*?\).*?>\n||gs;
    my $r = &parseresponse($o);
    my $pkg_name         = $r->{Name};
    my $pkg_version      = $r->{Version};
    my $pkg_release      = $r->{Release};
    my $pkg_distribution = $r->{Distribution};
    my $pkg_class        = $r->{Class};
    my $pkg_sigmd5       = $r->{SigMD5};
    my $pkg_breq         = $r->{BuildPreReq};

    #   update class information in database for use by build farm components
    &classify($pkg_name, $pkg_class);

    #   apply include/exclude class filter
    if (not $pkg_class =~ m/^(CORE|BASE|PLUS|PRIV|EVAL|JUNK)$/) {
        print STDERR "$progname:WARNING: skipping unknown $pkg_class\n";
        next SRCRPM;
    }
    if (defined @classcheck) {
        my $classhave  = $1;
        my $classpass  = 1 - $classcheck[0]->[1];
        foreach my $i (@classcheck) {
            if ($classhave eq $i->[0]) {
                $classpass = $i->[1];
            }
        }
        if (not $classpass) {
            print STDERR "$progname: skipping class $classhave due to class filter\n";
            next SRCRPM;
        }
    }

    #   query .spec inside source RPM package
    #   (because RPM has only single REQUIRE queries and applied to
    #   source RPMs they return the BuildPreReq and applied to the .spec
    #   they act as applied on a binary RPM and return the PreReq)
    #
    #   getting the spec through rpm2cpio is a very expensive operation
    #   in terms of IO and CPU burning. To avoid repetitive execution of
    #   this command the spec file is cached, speeding up the next run
    #   at the price of some hundret kilobytes disk space.

    #   read the cache (if any) and check whether it matches the src.rpm
    #   we work on. Becaues the OpenPKG naming scheme for CURRENT is
    #   build upon YYYYMMDD information it allows different packages
    #   with same name within a day. In order to detect changes anyway
    #   we check the md5.
    #
    sub cacheread ()
    {
        my ($cachefile) = @_;
        my $md5 = '';
        my $io = new IO::File "<$cachefile" or return $md5;
        $md5 .= $_ while (<$io>);
        $io->close();
        return $md5;
    }

    sub cachewrite ()
    {
        my ($cachefile, $md5) = @_;
        my $io = new IO::File ">$cachefile" or die "cannot open md5 cache '$cachefile'";
        $io->write($md5) or die "cannot write md5 into cache '$cachefile'";
        $io->close();
    }

    if ($pkg_sigmd5 ne &cacheread("$tmpdir/$srcrpm_name/md5")) {
        &verbose(sprintf "cache dirty (md5 mismatch): discard whole cache content, read original source and cache it");
        system("rm -rf $tmpdir/$srcrpm_name >/dev/null 2>&1 || true");
        mkdir("$tmpdir/$srcrpm_name", 0755) || die "cannot create temporary directory '$tmpdir/$srcrpm_name'";
        &runcmd("$rpm2cpio $srcdir/$srcrpm | (cd $tmpdir/$srcrpm_name && $cpio -idvmu $srcrpm_name.spec $srcrpm_name.sh >/dev/null 2>&1)");
        if (not -f "$tmpdir/$srcrpm_name/$srcrpm_name.spec") {
            print STDERR "$progname: failed to extract '$srcrpm_name.spec' from '$srcdir/$srcrpm' - skipping\n";
            next SRCRPM;
        }
        &cachewrite("$tmpdir/$srcrpm_name/md5", $pkg_sigmd5);
    }
    $q = "[M-PreReq:<%{REQUIRENAME} %|REQUIREFLAGS?{%{REQUIREFLAGS:depflags} %{REQUIREVERSION}}:{}|>\n]";
    $q .= "[M-Provides:<%{PROVIDENAME} %|PROVIDEFLAGS?{%{PROVIDEFLAGS:depflags} %{PROVIDEVERSION}}:{}|>\n]";
    $o = &cacheread("$tmpdir/$srcrpm_name/o");
    if ($o eq "") {
        &verbose(sprintf "cache dirty (o not cached): execute original query and cache it");
        $o = `cd $tmpdir/$srcrpm_name && $rpm -q --qf "$q" --define '_sourcedir .' --define '_specdir .' --specfile $srcrpm_name.spec 2>/dev/null`;
        $o =~ s|M-PreReq:<rpmlib\(.*?\).*?>\n||gs;
        &cachewrite("$tmpdir/$srcrpm_name/o", $o);
    }
    $r = &parseresponse($o);
    my $pkg_ireq     = $r->{PreReq};
    my $pkg_prov     = $r->{Provides};

    #    helper function for parsing the query outputs
    sub parseresponse {
        my ($o) = @_;
        my $r = {};
        $o =~ s|([SM])-([^:]+):<(.*?)>\n|parseline($r, $1, $2, $3, '')|egs;
        sub parseline {
            my ($r, $t, $k, $v) = @_;
            $v =~ s|^\s+||s;
            $v =~ s|\s+$||s;
            return if ($v =~ m|^\S+?::|);
            if ($t eq 'S') {     # single-value results
                $r->{$k} = $v;
            }
            elsif ($t eq 'M') {  # multi-value results
                $r->{$k} = [] if (not defined($r->{$k}));
                push(@{$r->{$k}}, $v);
            }
        }
        return $r;
    }

    #   consistency checks
    if ($pkg_name ne $srcrpm_name) {
        print STDERR "$progname:WARNING: inconsistent package name (file name '$srcrpm_name' != package name '$pkg_name')\n";
    }
    if ($pkg_version ne $srcrpm_version) {
        print STDERR "$progname:WARNING: inconsistent package version (file version '$srcrpm_version' != package version '$pkg_version')\n";
    }
    if ($pkg_release ne $srcrpm_release) {
        print STDERR "$progname:WARNING: inconsistent package release (file release '$srcrpm_release' != package release '$pkg_release')\n";
    }

    #   verbosity
    sub reqstr {
        my ($reqs) = @_;
        my $str = '';
        foreach my $req (@{$reqs}) {
            $str .= "\"$req\", ";
        }
        $str =~ s|,\s+$||s;
        return $str;
    }
    #&verbose("  build   requirement: ".&reqstr($pkg_breq));
    #&verbose("  install requirement: ".&reqstr($pkg_ireq));

    #   remember dependencies in graphs:
    #   this -> dep1, ..., this -> depN
    $DGB->{"$srcrpm_name-$srcrpm_version-$srcrpm_release"} = $pkg_breq;
    $DGI->{"$srcrpm_name-$srcrpm_version-$srcrpm_release"} = $pkg_ireq;

    #   remember all provides of current package
    #   virtual1 ->+ this, ..., virtualN ->+ this
    foreach my $prov (@{$pkg_prov}) {
        $prov =~ s|\s+\S+\s+\S+\s*$||s;
        $prov =~ s|\s+.*$||s;
        $PGV->{$prov} = [] if (not defined($PGV->{$prov}));
        push(@{$PGV->{$prov}}, $srcrpm_name);
    }

    #   remember package maximum version
    sub pkgislessthan {
        my ($a, $b) = @_;
        my ($a_v, $a_r) = ($a =~ m/^([^-]+)-(\d+\.\d+\.\d+|\d+\.\d{8}|\d{8})$/);
        my ($b_v, $b_r) = ($b =~ m/^([^-]+)-(\d+\.\d+\.\d+|\d+\.\d{8}|\d{8})$/);
        my @a_r = split(/\./, $a_r);
        my @b_r = split(/\./, $b_r);
        if ($#a_r > $#b_r) {
            return 1;
        }
        if ($#a_r < $#b_r) {
            return 0;
        }
        for (my $i = 0; $i < $#a_r; $i++) {
            if ($a_r[$i] < $b_r[$i]) {
                return 1;
            }
            if ($a_r[$i] > $b_r[$i]) {
                return 0;
            }
        }
        return 0;
    }
    if (not defined($PMV->{$srcrpm_name})) {
        $PMV->{$srcrpm_name} = "$srcrpm_version-$srcrpm_release";
    }
    elsif (&pkgislessthan($PMV->{$srcrpm_name}, "$srcrpm_version-$srcrpm_release")) {
        $PMV->{$srcrpm_name} = "$srcrpm_version-$srcrpm_release";
    }
}

#   debugging
if ($debug) {
    if (-f "$progname.db") {
        require "$progname.db";
    }
    else {
        sub dumphash {
            my ($n, $h) = @_;
            my $o;
            $o = Data::Dumper::Dumper($h);
            $o =~ s|^\s*\$VAR1|\$$n|s;
            return $o;
        }
        my $db = new IO::File ">$progname.db";
        $db->print(&dumphash("DGB", $DGB));
        $db->print(&dumphash("DGI", $DGI));
        $db->print(&dumphash("PGV", $PGV));
        $db->print(&dumphash("PMV", $PMV));
        $db->print("1;\n");
        $db->close;
    }
}

#
#   consistency checks
#

foreach my $pkg (sort(keys(%{$DGB}))) {
    foreach my $dep (@{$DGB->{$pkg}}) {
        my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
        if (not defined($PMV->{$name})) {
            print STDERR "$progname:WARNING: package '$pkg' has build dependency to unknown package '$name'\n";
        }
    }
}
foreach my $pkg (sort(keys(%{$DGI}))) {
    foreach my $dep (@{$DGI->{$pkg}}) {
        my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
        if (not defined($PMV->{$name})) {
            print STDERR "$progname:WARNING: package '$pkg' has install dependency to unknown package '$name'\n";
        }
    }
}

#
#   generate output Makefiles
#

&verbose("generating Makefile.rules");

my @MFTL = (); # Makefile target list
my $MF = '';   # Makefile contents

#   generate standard header
sub hdr {
    my ($line) = @_;
    my $hdr =
        "##\n" .
        "##  OpenPKG make(1) build specification for building/installing packages\n" .
        "##  ATTENTION: DO NOT EDIT THIS FILE, IT WAS AUTO-GENERATED BY $progname!\n" .
        "##\n" .
        "##  Copyright (c) 2000-2006 OpenPKG Foundation e.V. <http://openpkg.net/>\n" .
        "##  Copyright (c) 2000-2006 Ralf S. Engelschall <http://engelschall.com/>\n" .
        "##\n" .
        "##  Permission to use, copy, modify, and distribute this software for\n" .
        "##  any purpose with or without fee is hereby granted, provided that\n" .
        "##  the above copyright notice and this permission notice appear in all\n" .
        "##  copies.\n" .
        "##\n" .
        "##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED\n" .
        "##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n" .
        "##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.\n" .
        "##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR\n" .
        "##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,\n" .
        "##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT\n" .
        "##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF\n" .
        "##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n" .
        "##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,\n" .
        "##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT\n" .
        "##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF\n" .
        "##  SUCH DAMAGE.\n" .
        "##\n" .
        "##  $line\n" .
        "##\n" .
        "\n";
    return $hdr;
}

#   output standard header
$MF .= &hdr("Makefile.rules -- package build/installation rules");

#   output preamble
$MF .= "SHELL      = /bin/sh\n" .
       "\n" .
       "#   configuration (set by entrance Makefile)\n" .
       "HOSTNAME   = *\n" .
       "PLATFORM   = *\n" .
       "LOG        = *\n" .
       "SRC        = *\n" .
       "DST        = *\n" .
       "TMP        = *\n" .
       "RPM        = *\n" .
       "RPM_FLAGS  = *\n" .
       "EXEC_ROOT  = *\n" .
       "EXEC_USER  = *\n" .
       "EXEC_BFDB  = *\n" .
       "TERM_B     = *\n" .
       "TERM_N     = *\n" .
       "\n" .
       "#   transitive make(1) handling\n" .
       "MAKE_TRANS = MAKEFLAGS=\"\" \$(MAKE) \$(MFLAGS) -f Makefile.rules \\\n" .
       "             HOSTNAME=\"\$(HOSTNAME)\" PLATFORM=\"\$(PLATFORM)\" \\\n" .
       "             SSH_AUTH_SOCK=\"\$\$SSH_AUTH_SOCK\" \\\n" .
       "             TERM_B=\"\$(TERM_B)\" TERM_N=\"\$(TERM_N)\" \\\n" .
       "             LOG=\"\$(LOG)\" SRC=\"\$(SRC)\" DST=\"\$(DST)\" TMP=\"\$(TMP)\"\\\n" .
       "             RPM=\"\$(RPM)\" RPM_FLAGS=\"\$(RPM_FLAGS)\" \\\n" .
       "             EXEC_ROOT=\"\$(EXEC_ROOT)\" EXEC_USER=\"\$(EXEC_USER)\"\\\n" .
       "             EXEC_BFDB=\"\$(EXEC_BFDB)\"\n" .
       "\n" .
       "#   RPM read-only/read-write operations\n" .
       "RPM_RD     = MAKEFLAGS=\"\" \$(EXEC_USER) \$(RPM) \$(RPM_FLAGS)\n" .
       "RPM_RW     = MAKEFLAGS=\"\" \$(EXEC_ROOT) \$(RPM) \$(RPM_FLAGS)\n" .
       "\n" .
       "#   default target\n" .
       "default:\n" .
       "\t\@echo \"Do not run use Makefile.rules directly! Use Makefile instead.\"\n" .
       "\n";

#   output default target
$MF .= "#   default target\n" .
       "all: build\n" .
       "\n";
push(@MFTL, "all");

#   output standard targets
$MF .= "#   standard targets\n" .
       "build:         build-latest\n" .
       "install:     install-latest\n" .
       "uninstall: uninstall-latest\n" .
       "list:           list-latest\n" .
       "\n";
push(@MFTL, (qw(build install uninstall list)));

#   output
$MF .= "#   build all packages\n" .
       "build-all: \\\n";
push(@MFTL, "build-all");
foreach my $pkg (sort(keys(%{$DGB}))) {
    $MF .= "    build-$pkg \\\n";
}
$MF =~ s|\s+\\\n$|\n|s;
$MF .= "\n";

$MF .= "#   build all latest packages\n" .
       "build-latest: \\\n";
push(@MFTL, "build-latest");
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= "    build-$name \\\n";
}
$MF =~ s|\s+\\\n$|\n|s;
$MF .= "\n";

$MF .= "#   try to build all latest packages\n" .
       "try-build-latest:\n";
push(@MFTL, "try-build-latest");
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= "\t-\@\$(MAKE_TRANS) build-$name\n";
}
$MF .= "\n";

$MF .= "#   build individual latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= sprintf("%-30s %s\n", "build-$name:", "build-$name-".$PMV->{$name});
    push(@MFTL, "build-$name");
}
$MF .= "\n";

$MF .= "#   build individual-version latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    my $V = $PMV->{$name};
    $V =~ s/-([1-9][0-9]*)?[2-9][0-9]{7}$// || $V =~ s/-[1-9][0-9]*\.[0-9]+\.[0-9]+$//;
    $MF .= sprintf("%-30s %s\n", "build-$name-$V:", "build-$name-".$PMV->{$name});
    push(@MFTL, "build-$name-$V");
}
$MF .= "\n";

$MF .= "#   install all latest packages\n" .
       "install-latest: \\\n";
push(@MFTL, "install-latest");
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= "    install-$name \\\n";
}
$MF =~ s|\s+\\\n$|\n|s;
$MF .= "\n";

$MF .= "#   install individual latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= sprintf("%-30s %s\n", "install-$name:", "install-$name-".$PMV->{$name});
    push(@MFTL, "install-$name");
}
$MF .= "\n";

$MF .= "#   install individual-version latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    my $V = $PMV->{$name};
    $V =~ s/-([1-9][0-9]*)?[2-9][0-9]{7}$// || $V =~ s/-[1-9][0-9]*\.[0-9]+\.[0-9]+$//;
    $MF .= sprintf("%-30s %s\n", "install-$name-$V:", "install-$name-".$PMV->{$name});
    push(@MFTL, "install-$name-$V");
}
$MF .= "\n";

$MF .= "#   uninstall all latest packages\n" .
       "uninstall-latest: \\\n";
push(@MFTL, "uninstall-latest");
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= "    uninstall-$name \\\n";
}
$MF =~ s|\s+\\\n$|\n|s;
$MF .= "\n";

$MF .= "#   uninstall individual latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= sprintf("%-30s %s\n", "uninstall-$name:", "uninstall-$name-".$PMV->{$name});
    push(@MFTL, "uninstall-$name");
}
$MF .= "\n";

$MF .= "#   uninstall individual-version latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    my $V = $PMV->{$name};
    $V =~ s/-([1-9][0-9]*)?[2-9][0-9]{7}$// || $V =~ s/-[1-9][0-9]*\.[0-9]+\.[0-9]+$//;
    $MF .= sprintf("%-30s %s\n", "uninstall-$name-$V:", "uninstall-$name-".$PMV->{$name});
    push(@MFTL, "uninstall-$name-$V");
}
$MF .= "\n";

$MF .= "#   list all latest packages\n" .
       "list-latest: \\\n";
push(@MFTL, "list-latest");
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= "    list-$name \\\n";
}
$MF =~ s|\s+\\\n$|\n|s;
$MF .= "\n";

$MF .= "#   list individual latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    $MF .= sprintf("%-30s %s\n", "list-$name:", "list-$name-".$PMV->{$name});
    push(@MFTL, "list-$name");
}
$MF .= "\n";

$MF .= "#   list individual-version latest packages\n";
foreach my $name (sort(keys(%{$PMV}))) {
    my $V = $PMV->{$name};
    $V =~ s/-([1-9][0-9]*)?[2-9][0-9]{7}$// || $V =~ s/-[1-9][0-9]*\.[0-9]+\.[0-9]+$//;
    $MF .= sprintf("%-30s %s\n", "list-$name-$V:", "list-$name-".$PMV->{$name});
    push(@MFTL, "list-$name-$V");
}
$MF .= "\n";

$MF .= "#   list (print) individual-version-release\n";
foreach my $name (sort(keys(%{$PMV}))) {
    my $V = $PMV->{$name};
    $MF .= sprintf("%-30s\n\t\@echo %s\n", "list-$name-$V:", "$name-$V");
    push(@MFTL, "list-$name-$V");
}
$MF .= "\n";

#   "<name> [<op> <vers>]"
#   (<name>, <vers>)
sub fixversion {
    my ($PMV, $req) = @_;

    if ($req =~ m|^\S+$|) {
        return ($req, $PMV->{$req});
    }
    elsif ($req =~ m|^\s*(\S+)\s+(\S+)\s+(\S+)\s*$|) {
        my ($name, $op, $vers) = ($1, $2, $3);
        if ($op eq '=') {
            return ($name, $vers);
        }
        else {
            return ($name, $PMV->{$name});
        }
    }
    die "invalid requirement specification '$req'";
}

#   fixname
sub fixname {
    my ($PGV, $name) = @_;
    if (defined($PGV->{$name})) {
        return $PGV->{$name}->[0];
    }
    return $name;
}

#   generate build entry for a single package
foreach my $pkg (sort(keys(%{$DGB}))) {
    my ($pkg_name, $pkg_version, $pkg_release) =
        ($pkg =~ m/^(.+?)-([^-]+)-(\d+\.\d+\.\d+|\d+\.\d{8}|\d{8})$/);

    $MF .= "#    build: $pkg\n";
    #    standard packages
    $MF .= "build-$pkg: \$(DST)/$pkg.\$(PLATFORM).rpm\n";
    push(@MFTL, "build-$pkg");
    $MF .= "\$(DST)/$pkg.\$(PLATFORM).rpm: \$(SRC)/$pkg.src.rpm\n";
    push(@MFTL, "\$(DST)/$pkg.\$(PLATFORM).rpm");

    #   generate dependency information
    my $il = '';
    my $unique = {};
    foreach my $dep (@{$DGB->{$pkg}}) {
        my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
        $il .= " build-$name-$vers" if (not defined($unique->{"$name-$vers"}));
        $unique->{"$name-$vers"} = 1;
    }
    $MF .= "\t\@\$(MAKE_TRANS)".$il."\n" if ($il ne '');

    #   generate build prolog (installation of dependencies)
    $il = '';
    foreach my $dep (@{$DGB->{$pkg}}) {
        my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
        $il .= " install-$name-$vers";
    }
    $MF .= "\t\@\$(MAKE_TRANS)".$il."\n" if ($il ne '');

    #   generate build command
    $MF .= "\t-\@echo \"\$(TERM_B):::: Building $pkg ::::\$(TERM_N)\"\n";
    $MF .= "\t-\$(EXEC_BFDB) -i \$(HOSTNAME)-\$(PLATFORM) -p $pkg begin\n";
    $MF .= "\t-rm -rf \$(TMP)/SRC/* \$(TMP)/TMP/* 2>/dev/null\n";
    $MF .= "\t\@cd \$(SRC) && (echo \"Build Start: \`date\`\"; " .
           "\$(RPM_RD) --rebuild $pkg.src.rpm 2>&1; " .
           "echo \"Build End: \`date\`\") | " .
           "tee \$(LOG)/$pkg.log.\$(HOSTNAME).\$(PLATFORM)\n";
    $MF .= "\t-color=\"red\"; dst=\$(DST)/$pkg.\$(PLATFORM).rpm; if [ -f \$\$dst ]; then \$(RPM_RD) --checksig \$\$dst; if [ \$\$? -ne 0 ]; then rm \$\$dst; else color=\"green\"; fi; fi; " .
           "\$(EXEC_BFDB) -i \$(HOSTNAME)-\$(PLATFORM) -p $pkg -c \$\$color end; " .
           "[ \$\$color = \"red\" ] && rm $pkg.src.rpm; [ \$\$color = \"green\" ]\n";

    #   generate build epilog (installation of dependencies)
    if (not $nouninstall) {
        $il = '';
        $unique = {};
        foreach my $dep (reverse(@{$DGB->{$pkg}})) {
            my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
            $il .= " uninstall-$name-$vers" if (not defined($unique->{"$name-$vers"}));
            $unique->{"$name-$vers"} = 1;
        }
        $MF .= "\t\@\$(MAKE_TRANS)".$il."\n" if ($il ne '');
    }
    $MF .= "\n";
}

#   generate install/uninstall entries for a single package
foreach my $pkg (sort(keys(%{$DGI}))) {
    my ($pkg_name, $pkg_version, $pkg_release) =
        ($pkg =~ m/^(.+?)-([^-]+)-(\d+\.\d+\.\d+|\d+\.\d{8}|\d{8})$/);

    #   generate install target
    $MF .= "#    install: $pkg\n";
    $MF .= "install-$pkg: build-$pkg\n";
    push(@MFTL, "install-$pkg");

    #   generate dependency information
    my $il = '';
    my $unique = {};
    foreach my $dep (@{$DGI->{$pkg}}) {
        my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
        $il .= " install-$name-$vers" if (not defined($unique->{"$name-$vers"}));
        $unique->{"$name-$vers"} = 1;
    }
    $MF .= "\t\@\$(MAKE_TRANS)".$il."\n" if ($il ne '');

    $MF .= "\t-\@echo \"\$(TERM_B):::: Installing $pkg ::::\$(TERM_N)\"\n";
    $MF .= "\t\$(RPM_RD) -q $pkg >/dev/null || \$(RPM_RW) -Uh --oldpackage \$(DST)/$pkg.\$(PLATFORM).rpm\n";

    $MF .= "\n";

    #   generate uninstall target
    $MF .= "#    uninstall: $pkg\n";
    $MF .= "uninstall-$pkg:\n";
    push(@MFTL, "uninstall-$pkg");

    #   generate uninstall operation
    if (not $nouninstall and not $pkg_name eq 'openpkg') {
        $MF .= "\t-\@echo \"\$(TERM_B):::: Uninstalling $pkg ::::\$(TERM_N)\"\n";
        $MF .= "\t\$(RPM_RD) -q $pkg >/dev/null && \$(RPM_RW) -e --noscripts --allmatches $pkg >/dev/null 2>&1 || true\n";
    }

    $il = '';
    $unique = {};
    foreach my $dep (reverse(@{$DGI->{$pkg}})) {
        my ($name, $vers) = &fixversion($PMV, &fixname($PGV, $dep));
        $il .= " uninstall-$name-$vers" if (not defined($unique->{"$name-$vers"}));
        $unique->{"$name-$vers"} = 1;
    }
    $MF .= "\t\@\$(MAKE_TRANS)".$il."\n" if ($il ne '');

    $MF .= "\n";
}

&verbose("writing Makefile.rules");

my $mf = new IO::File ">$outdir/Makefile.rules" || die "unable to create output '$outdir/Makefile.rules'";
$mf->print($MF);
$mf->close;

&verbose("generating Makefile");

$MF = &hdr("Makefile -- end-user entrance targets");

$MF .= "SHELL      = /bin/sh\n" .
       "\n";

$MF .= "#   configuration: OpenPKG hierarchy\n" .
       "PREFIX     = $prefix\n" .
       "HOSTNAME   = auto\n" .
       "PLATFORM   = auto\n" .
       "\n" .
       "#   configuration: OpenPKG package areas\n" .
       "TMP        = /tmp/openpkg\n" .
       "LOG        = .\n" .
       "SRC        = .\n" .
       "DST        = .\n" .
       "\n" .
       "#   configuration: OpenPKG RPM tool\n" .
       "RPM        = \$(PREFIX)/bin/openpkg rpm\n" .
       "RPM_FLAGS  = \n" .
       "\n" .
       "#   configuration: user/root command execution\n" .
       "EXEC_HOST  = localhost\n" .
       "#EXEC_ROOT  = ssh -t -x -q root\@\$(EXEC_HOST) cd \$\$PWD \\&\\&\n" .
       "EXEC_ROOT  = sudo\n" .
       "EXEC_USER  = \n" .
       "EXEC_BFDB  = openpkg bf-db\n" .
       "\n" .
       "#   configuration: command execution environment\n" .
       "ENV_PATH   = auto\n" .
       "ENV_LDLP   = auto\n" .
       "\n";

my $TL = '';
foreach my $tl (@MFTL) {
    $TL .= "$tl \\\n";
}
$MF .= "#   all-in-one transition target\n" .
       "${TL}".
       "::\n" .
       "\t\@PLATFORM=\"\$(PLATFORM)\"; \\\n" .
       "\tif [ \".\$\$PLATFORM\" = .auto ]; then \\\n" .
       "\t    PLATFORM=\"\`\$(RPM) --eval '%{l_platform -p}-%{l_tag}'\`\"; \\\n" .
       "\tfi; \\\n" .
       "\tHOSTNAME=\"\$(HOSTNAME)\"; \\\n" .
       "\tif [ \".\$\$HOSTNAME\" = .auto ]; then \\\n" .
       "\t    HOSTNAME=\"\`hostname | sed 's;\\..*\$\$;;'\`\"; \\\n" .
       "\tfi; \\\n" .
       "\tENV_PATH=\"\$(ENV_PATH)\"; \\\n" .
       "\tENV_LDLP=\"\$(ENV_LDLP)\"; \\\n" .
       "\tSYSTEM=\"`(uname -s) 2>/dev/null || echo unknown`\"; \\\n" .
       "\tif [ \".\$\$ENV_PATH\" = .auto ]; then \\\n" .
       "\t    case \$\$SYSTEM in \\\n" .
       "\t        FreeBSD ) ENV_PATH=\"/bin:/sbin:/usr/bin:/usr/sbin:/usr/opkg/bin\" ;; \\\n" .
       "\t        Linux   ) ENV_PATH=\"/bin:/sbin:/usr/bin:/usr/sbin:/usr/opkg/bin\" ;; \\\n" .
       "\t        SunOS   ) ENV_PATH=\"/bin:/sbin:/usr/bin:/usr/sbin:/usr/ccs/bin:/usr/opkg/bin\" ;; \\\n" .
       "\t        *       ) ENV_PATH=\"/bin:/sbin:/usr/bin:/usr/sbin:/usr/opkg/bin\" ;; \\\n" .
       "\t    esac; \\\n" .
       "\tfi; \\\n" .
       "\tif [ \".\$\$ENV_LDLP\" = .auto ]; then \\\n" .
       "\t    case \$\$SYSTEM in \\\n" .
       "\t        FreeBSD ) ENV_LDLP=\"/usr/lib\" ;; \\\n" .
       "\t        Linux   ) ENV_LDLP=\"/lib:/usr/lib\" ;; \\\n" .
       "\t        SunOS   ) ENV_LDLP=\"/usr/lib:/usr/ccs/lib\" ;; \\\n" .
       "\t        *       ) ENV_LDLP=\"/usr/lib\" ;; \\\n" .
       "\t    esac; \\\n" .
       "\tfi; \\\n" .
       "\tTERM_B=\"\"; TERM_N=\"\"; \\\n" .
       "\tcase \$\${TERM:-vt100} in \\\n" .
       "\t    xterm|xterm*|vt220|vt220*) \\\n" .
       "\t        TERM_B=`awk 'BEGIN { printf(\"%c%c%c%c\", 27, 91, 49, 109); }' </dev/null 2>/dev/null`; \\\n" .
       "\t        TERM_N=`awk 'BEGIN { printf(\"%c%c%c\", 27, 91, 109); }' </dev/null 2>/dev/null` \\\n" .
       "\t        ;; \\\n" .
       "\t    vt100|vt100*) \\\n" .
       "\t        TERM_B=`awk 'BEGIN { printf(\"%c%c%c%c%c%c\", 27, 91, 49, 109, 0, 0); }' </dev/null 2>/dev/null`; \\\n" .
       "\t        TERM_N=`awk 'BEGIN { printf(\"%c%c%c%c%c\", 27, 91, 109, 0, 0); }' </dev/null 2>/dev/null` \\\n" .
       "\t        ;; \\\n" .
       "\tesac; \\\n" .
       "\tumask 022; \\\n" .
       "\tif [ ! -d \$(TMP) ]; then \\\n" .
       "\t    mkdir \$(TMP); \\\n" .
       "\tfi; \\\n" .
       "\tif [ ! -d \$(TMP)/SRC ]; then \\\n" .
       "\t    mkdir \$(TMP)/SRC; \\\n" .
       "\tfi; \\\n" .
       "\tif [ ! -d \$(TMP)/TMP ]; then \\\n" .
       "\t    mkdir \$(TMP)/TMP; \\\n" .
       "\tfi; \\\n" .
       "\trm -rf \$(TMP)/SRC/* 2>/dev/null || true; \\\n" .
       "\trm -rf \$(TMP)/TMP/* 2>/dev/null || true; \\\n" .
       "\tLOG=\"\`cd \$\(LOG) && pwd\`\"; \\\n" .
       "\tSRC=\"\`cd \$\(SRC) && pwd\`\"; \\\n" .
       "\tDST=\"\`cd \$\(DST) && pwd\`\"; \\\n" .
       "\tTMP=\"\`cd \$\(TMP) && pwd\`\"; \\\n" .
       "\t( echo \"%_sourcedir \$\$TMP/SRC/%{name}\"; \\\n" .
       "\t  echo \"%_specdir   \$\$TMP/SRC/%{name}\"; \\\n" .
       "\t  echo \"%_builddir  \$\$TMP/TMP\"; \\\n" .
       "\t  echo \"%_tmppath   \$\$TMP/TMP\"; \\\n" .
       "\t  echo \"%_rpmdir    \$\$DST\"; \\\n" .
       "\t) >\$(TMP)/.rpmmacros; \\\n" .
       "\trm -f Makefile.log.\$\$PLATFORM 2>/dev/null || true; \\\n" .
       "\tenv -i \\\n" .
       "\t    HOME=\"\$(TMP)\" \\\n" .
       "\t    SHELL=\"\$(SHELL)\" TERM=\"\$\$TERM\" \\\n" .
       "\t    TMPDIR=\"\$(TMP)\" TEMPDIR=\"\$(TMP)\" \\\n" .
       "\t    USER=\"\$\$USER\" LOGNAME=\"\$\$LOGNAME\" \\\n" .
       "\t    PATH=\"\$\$ENV_PATH\" LD_LIBRARY_PATH=\"\$\$ENV_LDLP\" \\\n" .
       "\t    \$(MAKE) \$(MFLAGS) -f Makefile.rules \\\n" .
       "\t        HOSTNAME=\"\$\$HOSTNAME\" PLATFORM=\"\$\$PLATFORM\" \\\n" .
       "\t        SSH_AUTH_SOCK=\"\$\$SSH_AUTH_SOCK\" \\\n" .
       "\t        TERM_B=\"\$\$TERM_B\" TERM_N=\"\$\$TERM_N\" \\\n" .
       "\t        LOG=\"\$\$LOG\" SRC=\"\$\$SRC\" DST=\"\$\$DST\" TMP=\"\$\$TMP\" \\\n" .
       "\t        RPM=\"\$(RPM)\" RPM_FLAGS=\"\$(RPM_FLAGS)\" \\\n" .
       "\t        EXEC_ROOT=\"\$(EXEC_ROOT)\" EXEC_USER=\"\$(EXEC_USER)\" \\\n" .
       "\t        EXEC_BFDB=\"\$(EXEC_BFDB)\" \\\n" .
       "\t        \$\@ 2>&1 | tee Makefile.log.\$\$PLATFORM\n" .
       "\n";

$MF .= "#   cleanups\n" .
       "clean:\n" .
       "\trm -rf \$(TMP) >/dev/null 2>&1 || true\n" .
       "\trm -f Makefile.log.* >/dev/null 2>&1 || true\n" .
       "\n";

&verbose("writing Makefile");

$mf = new IO::File ">$outdir/Makefile" || die "unable to create output '$outdir/Makefile'";
$mf->print($MF);
$mf->close;

#
#   Die gracefully...
#

&cleanup_perform();
exit(0);

#   this code reading config is shared between bf-mk.pl, bf-db.pl,
#   bf-ui.pl and src2make.pl. It is a candidate to be moved into a perl
#   module shared between the three as soon as such a module is created.
#
sub subst ($$)
{
    my ($cfg, $txt) = @_;
    while ($txt =~ m|%{([^}]+)\.([^}]+)}|) {
        my ($section, $k) = ($1, $2);
        my $v = $cfg->{$section}->{$k};
        if (defined $v) {
            $txt =~ s|%{$section.$k}|$v|g;
        }
        else {
            die "undefined variable %{$section.$k}";
        }
    }
    return $txt;
}

sub readconfig($)
{
    my ($cfgfile) = @_;
    my $cfg = {};
    my $osspcfg = new OSSP::cfg::simple;

    die "no config file specified" unless (defined $cfgfile);
    my $io = new IO::File "<$cfgfile";
    die "cannot open \"$cfgfile\" for reading ($!)" if(not defined $io);
    my $txt = ''; $txt .= $_ while (<$io>);
    $io->close();

    $osspcfg->parse($txt) || die "error parsing config file";
    my $tree = $osspcfg->unpack(-index => '.*') || die "error unpacking config file";

    foreach my $section (qw(master shared driver slaves)) {
        foreach my $k (keys %{$tree->{$section}->[1]}) {
            my $v = $tree->{$section}->[1]->{$k}->[1];
            $v =~ s|%{([^}\.]+)}|%{$section.\1}|g; #expand shortcuts with section omitted
            $cfg->{$section}->{$k} = $v;
            &verbose("section=$section, k=$k, v=$v");
        }
    }
    foreach my $section (keys %{$cfg}) {
        foreach my $k (keys %{$cfg->{$section}}) {
            $cfg->{$section}->{$k} = &subst($cfg, $cfg->{$section}->{$k});
        }
    }
    return($cfg);
}

#   open/create database and create all missing tables on the fly
sub dbinit ($)
{
    my ($dbfile) = @_;
    my $dbh = DBI->connect("dbi:SQLite:dbname=$dbfile", "", "", { RaiseError => 0, AutoCommit => 1 });
    if (not defined $dbh) {
        print STDERR "ERROR:$0: \"cannot open database file \"$dbfile\". DataBase error: ".$dbh->errstr."\n";
        exit(2);
    }

    my $tables = $dbh->selectall_hashref("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;", "name");

    unless ($tables->{class}) {
        printf STDERR "DEBUG:$0: creating \"class\" table in database file \"%s\"\n", $dbfile if ($verbose);
        my $sql = '';
        $sql .= "CREATE TABLE class";                          # field description                  | example
        $sql .= "(\n";                                         # -----------------------------------+-----------
        $sql .= "    cl_package_name    TEXT NOT NULL,\n";     # 1st part of $pkg in Makefile.rules | a2ps
        $sql .= "    cl_package_class   TEXT,\n";              # CORE|BASE|PLUS|EVAL|JUNK           | CORE
        $sql .= "PRIMARY KEY ( cl_package_name ) ON CONFLICT REPLACE\n";
        $sql .= ");\n";
        $dbh->do($sql);
    };

    return $dbh;
}

sub classify ($$)
{
    my ($name, $class) = @_;
    my ($sth, $rv, $retries);
    printf STDERR "DEBUG:$0: classify(%s, %s);\n", $name, $class if ($debug);

    $sth = $dbh->prepare(
        "INSERT INTO class ( " .
        "    cl_package_name, cl_package_class )" .
        "VALUES ( ?, ? );"
    );
    $retries = $lockretries; do {
        $rv = $sth->execute(
            $name, $class
            );
    } while (not $rv and $dbh->errstr =~ m|lock|i and sleep(1) and $retries-- >= 1);
    printf STDERR "DEBUG:$0: countdown from %d to %d\n", $lockretries, $retries if($debug or $retries != $lockretries);
    if (not $rv) {
        print STDERR "ERROR:$0: DataBase error: ".$dbh->errstr."\n";
        exit(2);
    }
}

